% Example for robust eigen classification: the minimal ratio is
% not the smallest eigenvalue of the 1st NEP.

clear;  close all;
DOCONTOUR = true; % do contour plot if set true

addpath ./rgec/
load ex1.mat; 

AA= [A,-ones(length(A),1)]; AA = AA'*AA;
BB= [B,-ones(length(B),1)]; BB = BB'*BB;
[VV0,EE0] = eig(AA,BB);

EvalTG = @(z) evalTG(z, A, SIGA);
EvalTH = @(z) evalTH(z, B, SIGB);

z1 = VV0(:,1); z0 = z1/norm(z1);

% ++++++++++++++++++++++++++++++
% --- Compute the optimal point
% ++++++++++++++++++++++++++++++
% Run simple iteration
[z1, r, R] = scfitr(EvalTG, EvalTH, z0);
% Run SCF to shifted 1st NEP
[z2, r2, R2, ~, resid] = scfsls(EvalTG, EvalTH, z0, true, 1.0E-8);
% Run SCF to 2nd NEP 
[z3, r3, R3, ~, resid2] = scfls(EvalTG, EvalTH, z0, false, 1.0E-10);

z = z3;

%Refine the solution z by MATLAB fminuc
%options.TolFun = 1e-16;
%[z, rho] = fminunc(@(z) evalrho(z, A, B, SIGA, SIGB), z2); %, options);
%z = z/z(3); % normalize


% ++++++++++++++++++++++++++++++
% --- Convergence history 
% ++++++++++++++++++++++++++++++
figure
plot(r,'--+','LineWidth',2,'MarkerSize',8);
hold on;
plot(r2,'--xk','LineWidth',2,'MarkerSize',8);
plot(r3,'--or','LineWidth',2,'MarkerSize',8);
xlabel('iterations','Fontsize',10);
ylabel('RQ values \rho(z_k)','Fontsize',10);
h_legend=legend('simple iter', 'Alg.1', 'Alg.2');
set(h_legend,'FontSize',10);
xlim([0,20]);
ylim([0,3]);

figure
semilogy(abs(r2-r2(end)),'--xk','LineWidth',2,'MarkerSize',8);
hold on;
semilogy(abs(r3-r3(end)),'--or','LineWidth',2,'MarkerSize',8);
xlabel('iterations','Fontsize',10);
ylabel('|\rho(z_k) -\rho(z_*)|','Fontsize',10);
h_legend=legend('Alg.1', 'Alg.2');
xlim([0,20]);
set(h_legend,'FontSize',10);

% ++++++++++++++++++++++++++++++
% --- Display eigenvalue order at z*, computed by 2nd NEP+SCF
% ++++++++++++++++++++++++++++++

% Eig of first order condition
[G, TTA] = evalG(z3, A, SIGA);
[H, TTB] = evalH(z3, B, SIGB);
[V0, E0] = eig(G, H); V0 = V0/diag(V0(3,:));

% Eig of second order condition
[G, TG] = evalTG(z3, A, SIGA);
GG = G + TG;
[H, TH] = evalTH(z3, B, SIGB);
HH = H + TH;
[VV, EE] = eig(GG, HH); VV =VV/diag(VV(3,:));

disp('----------- PROFILE -----------')
fprintf(' Optimal ratio by SCF to 1st NEP , and residual |Gz-rho*Hz|/|z|:\n\t %16.15f,\t%2.1E \n', r2(end),  resid(end));
fprintf('\n Optimal ratio by SCF to 2nd NEP, and residual |Gz-rho*Hz|/|z|:\n\t %16.15f,\t %2.1E\n', r3(end), resid2(end));
%
fprintf('\n-------- First order NEP\n')
disp(' Eigenvalues   ||  eigvecs of Gx=eHx   ||  optimal z');
fprintf(' %+6.5f    ||  %+6.5f |  %+6.5f |  %+6.5f  ||  %+6.5f\n', [diag(E0),V0, z2/z2(end)]');

fprintf('\n-------- Second order NEP\n')
disp(' Eigenvalues   ||  eigvecs of Gx=eHx   ||  optimal z');
fprintf(' %+6.5f    ||  %+6.5f |  %+6.5f |  %+6.5f  ||  %+6.5f\n', [diag(EE),VV, z3/z3(3)]');


% ++++++++++++++++++++++++++++++
% --- Plotting the classification lines
% ++++++++++++++++++++++++++++++
figure

% data points of A, B
plot(A(:,1),A(:,2),'ok','MarkerSize',6,'LineWidth',2);
hold on;
plot(B(:,1),B(:,2),'+m','MarkerSize',6,'LineWidth',2);
axis equal;

% classification plane
zz = R2(:,1); 
xa = -2; xb = 4.5;
plot([xa,xb],[(zz(3)-xa*zz(1))/zz(2),(zz(3)-xb*zz(1))/zz(2)],'--k', 'linewidth',2);
plot([xa,xb],[(z3(3)-xa*z3(1))/z3(2),(z3(3)-xb*z3(1))/z3(2)],'-k','linewidth',2);

h_legend=legend('class A','class B','eig plane','robust plane', 'Location','NorthEast');
xlabel('x','FontSize',10); ylabel('y','rot',0,'FontSize',10);

% uncertainty ellipsoids
theta = 0:.1:2*pi;
alpha = SIGA(1,1)^-.5; beta = SIGA(2,2)^-.5;
for i = 1: m
    X = alpha*cos(theta);
    Y = beta*sin(theta);
    %-------------------------%
    Z = T1*[X;Y];   % transformation, T1 = I
    %-------------------------%
    X = A(i,1)+ Z(1,:);
    Y = A(i,2)+ Z(2,:);
    plot(X,Y,'-k','LineWidth',1);
end 
alpha = SIGB(1,1)^-.5; beta = SIGB(2,2)^-.5;
for i = 1: p
    X = alpha*cos(theta);
    Y = beta*sin(theta);
    %-------------------------%
    Z = T2*[X;Y];   % transformation, T2 = I
    %-------------------------%
    X = B(i,1)+ Z(1,:);
    Y = B(i,2)+ Z(2,:);
    plot(X,Y,'-m','LineWidth',1);
end
xlim([xa, xb]);
ylim([-1, 6]);

if ~DOCONTOUR, return, end

% ++++++++++++++++++++++++++++++
% --- Contour plot of rho close to z optimal
% ++++++++++++++++++++++++++++++
lx = -2; ux = 1;
ly = -4; uy = 3;
nsamples = 30;
zz = z/z(3);
xx = zz(1)+linspace(lx, ux, nsamples);
yy = zz(2)+linspace(ly, uy, nsamples);
[XX,YY] = meshgrid(xx, yy);
RHO = zeros(nsamples, nsamples);
% parfor i = 1 : nsamples
for i = 1 : nsamples    
    for j = 1 : nsamples
        zz = [xx(i);yy(j); 1];
        RHO(j,i) = evalrho(zz, A, B, SIGA, SIGB);
    end
end

figure;
h = meshc_m(XX, YY, RHO); 
set(h(2:end), 'LineWidth', 2);
colormap('default')
xlabel('z_1')
ylabel('z_2')
zlabel('$\rho(z)$','Interpreter', 'latex', 'Rotation', 0)
hold on; 
plot3(z(1)/z(3), z(2)/z(3), -10, '+k', 'linewidth', 3, 'markersize', 8);


