% Example for robust eigen classification: Pima-Indian data set
% Testing problem:  breast pima
%                   perturbation added to data

%% UNCOMMENT TO USE parallel computation
% matlabpool close;
% matlabpool open allcluster;

addpath ./rgec/

% ++++++++++++++++++++++++++++++
% ---  DATA SETTING UP
% ++++++++++++++++++++++++++++++

clear; close all; clc;
rng('default'); % for reproducibility

load pima.mat;
% pima is an n-by-9 matrix
% n is the number of data entries
% 9 is the number of attributes
% Each Attribute: (all numeric-valued)
%   1. Number of times pregnant
%   2. Plasma glucose concentration a 2 hours in an oral glucose tolerance test
%   3. Diastolic blood pressure (mm Hg)
%   4. Triceps skin fold thickness (mm)
%   5. 2-Hour serum insulin (mu U/ml)
%   6. Body mass index (weight in kg/(height in m)^2)
%   7. Diabetes pedigree function
%   8. Age (years)
%   9. Class variable (0 or 1)

% Solvers for robust classifier SOLVERS = {'scfls', 'scfsls', 'scfitr'};
SOLVERS = {'scfls'};

tstr = .7;  % testing set rate, training set rate = 1 - tstr
            % follow 70/30 rule, 70% training, 30% validation
N_SAMPLES = 10;     % number of repeated experiments
TST_SAMPLES = 500;  % number of perturbed testing, inner iterations

BETA = 10; %0.1; 1; 2;
TST_SCALE_RATE = 1/BETA;% The ellipsoids used for perturbed the testing data
                       % is SA*TST_SCALE_RATE*rand and SB*TST_SCALE_RATE*rand
                       % To test sensitivity of robust classifier, when the
                       % ellipsoids used for training is not "exact":
                       % TST_SCALE_RATE > 1, trainig SA, SB, over estimated
                       % TST_SCALE_RATE < 1, trainig SA, SB, under estimated
                       %
                       % Expected:
                       % TST_SCALE_RATE : large   RGEC : conservative (low variance, low accurcy)
                       % TST_SCALE_RATE : small   RGEC : aggressive (high variance)


n = length(pima);   % #of entries
nf= size(pima,2)-1; % #of features
e = mean(pima(:,1:nf)); % mean

% Define ellipsoids 
SA = eye(nf)/diag(e.^2)/(.5^2); 
SA(end,end) = SA(end,end)/(.002^2);    % the age is not perturbed
SA(1,1) = SA(1,1)/(.002^2);            % the number of pregnant
SB = eye(nf)/diag(e.^2)/(.5^2); 
SB(end,end) = SB(end,end)/(.002^2);    % the age is not perturbed
SB(1,1) = SB(1,1)/(.002^2);            % the number of pregnant

nn = round(tstr*n); %number of training instances

frr1 = []; frr2 = []; frr3 = []; % miss-classification rate
nita2 = []; nitb2 =[];  % DSCF
nita3 = []; nitb3 =[];

for kkk = 1:length(SOLVERS)
    NITA{kkk} = []; NITB{kkk} =[];
    RA{kkk} = []; RB{kkk} =[];
    ResA{kkk} = []; ResB{kkk} =[];
end

% ++++++++++++++++++++++++++++++
% --- MAIN LOOP: REPEAT EXPEREMENTS
% ++++++++++++++++++++++++++++++

fprintf(1, ' ----- Progress %3d/%3d', 0, N_SAMPLES); % progress bar

for kk = 1:N_SAMPLES 
    %
    % Generate training (ratio=tstr) and testing data
    %
    pima0 = pima(randperm(n),:);    % apply random row permulation to database
    T = pima0(1:nn,:);      % training data
    C = pima0(nn+1:end,:);  % tsting data
    D = sortrows(T,nf+1);   % Sort the training data according to the class label
    nA = length(find(D(:,end)==0)); % #data entry in class A, labeled 0
    nB = length(D) - nA;    % #data entry in class B, labeled 1
    A = D(1:nA,:);
    B = D(nA+1:end,:);
    SIGA = kron(ones(1,nA),SA);
    SIGB = kron(ones(1,nB),SB);

    EvalTG = @(z) evalTG(z, A(:,1:nf), SIGA);
    EvalTH = @(z) evalTH(z, B(:,1:nf), SIGB);
    EvalTG2 = @(z) evalTG(z, B(:,1:nf), SIGB);
    EvalTH2 = @(z) evalTH(z, A(:,1:nf), SIGA);

    %------------------------------------
    %   Compute unperturbed planes: GEC
    %------------------------------------
    AA = [A(:,1:nf),-ones(nA,1)]; AA = AA'*AA;
    BB = [B(:,1:nf),-ones(nB,1)]; BB = BB'*BB;
    [V,E] = eig( AA, BB );
    [M, idx] = max(diag(E)); 
    lbb = V(:,idx); 
    [M, idx] = min(diag(E));
    laa = V(:,idx); 

    for kkk = 1:length(SOLVERS)
        rbstsolver = eval(['@', SOLVERS{kkk}]);
        %------------------------------------
        %   Compute plane A
        %------------------------------------
        [la3, r3, R3, nita31, res3] = rbstsolver(EvalTG,EvalTH,laa); 
        %------------------------------------
        %   Compute plane B
        %------------------------------------
        [lb3, rb3, Rb3, nitb31, resb3] = rbstsolver(EvalTG2,EvalTH2,lbb); 

        % record classifier 
        LA{kkk} = la3; 
        LB{kkk} = lb3;

        % history
        NITA{kkk} = [NITA{kkk}, nita31]; 
        NITB{kkk} = [NITB{kkk}, nitb31];
        RA{kkk} = [RA{kkk}, r3(end)]; RB{kkk} =[RB{kkk}, rb3(end)];
        ResA{kkk} = [ResA{kkk}, res3(end)]; ResB{kkk} =[ResB{kkk}, resb3(end)];
    end

    %------------------------------------
    %   Accuracy test
    %------------------------------------
    % REPEAT NN TIMES
    xrat = []; % missing rate
    NN = TST_SAMPLES; % sample times

    % normalize classifier
    LAA = laa/norm(laa(1:nf));
    LBB = lbb/norm(lbb(1:nf));
    for kkk = 1:length(SOLVERS)
        la3 = LA{kkk}; lb3 = LB{kkk};
        LAA = [LAA, la3/norm(la3(1:nf))];
        LBB = [LBB, lb3/norm(lb3(1:nf))];
    end

    for i = 1: NN
        % Add perturbation to testing data
        C = sortrows(C,nf+1);
        nna = length(find(C(:,end)==0)); % number of class A entry
        nnb = length(C) - nna; % number of class B entry
        Ra = chol(SA*TST_SCALE_RATE);
        DA = (randn(nna,nf))/Ra'; % Gaussian distributed data A perturbation
        Rb = chol(SB*TST_SCALE_RATE);
        DB = (randn(nnb,nf))/Rb'; % Gaussian distributed data B perturbation
        C(:,1:nf) = C(:,1:nf) + [DA;DB]; % perturb testing instances
        %
        % check rate of misclassification
        %
        CC = C(:,end); % class label of testing sets
        dA = [C(:,1:nf),-ones(length(C),1)]*LAA; %[laa,la,la2,la3]; % distance to class A
        dB = [C(:,1:nf),-ones(length(C),1)]*LBB; %[lbb,lb,lb2,lb3]; % distance to class B
        I = abs(dA)>abs(dB); %I(i) =1, belong to class B; 0 to A;
        I = I - CC*ones(1,1+length(SOLVERS)); %ones(1,4);
        %   xrat = [xrat; mr0, mr, mr2, mr3]; 
        xratt = [];
        for kkk = 1:length(SOLVERS) + 1
            mrr = length(find(I(:,kkk))) / length(C); 
            xratt = [xratt, mrr];
        end
        xrat = [xrat; xratt]; 
    end;

    wrat = max(xrat); % worst rate, highest miss-classification
    mrat = mean(xrat); % mean rate
    brat = min(xrat); % best rate
    frr1 = [frr1;brat];
    frr2 = [frr2;mrat];
    frr3 = [frr3;wrat];

    fprintf(1, [repmat('\b',1,7),'%3d/%3d'], kk, N_SAMPLES); % progress bar
end

fprintf(1, '\n'); % progress bar

xx = 1:size(frr1,1);
[frr2, IDX] = sortrows(frr2,1); % sort miss-classification by GEC
frr1 = frr1(IDX,:);
frr3 = frr3(IDX,:);

% ++++++++++++++++++++++++++++++
% --- display
% ++++++++++++++++++++++++++++++
figure
errorbar(xx, 1-frr2(:,1), abs(frr3(:,1)-frr2(:,1)), abs(frr2(:,1)-frr1(:,1)), ...
    'sm', 'linewidth', 1, 'markerfacecolor','m','DisplayName', 'GEC',...
    'MarkerSize', 8, 'CapSize', 5);
hold on;

if length(SOLVERS) ~= 1 % TESTING MODE
    corder = lines(length(SOLVERS));
    SOLVERS_LABEL = SOLVERS;
else 
    corder = 'b';
    SOLVERS_LABEL = {'RGEC'};
end

for kkk = 1:length(SOLVERS)
    errorbar(xx+0.2, 1-frr2(:,1+kkk), abs(frr3(:,1+kkk)-frr2(:,1+kkk)), ...
        abs(frr2(:,1+kkk)-frr1(:,1+kkk)),...
        'o', 'linewidth', 1, 'color', corder(kkk,:), ...
        'markerfacecolor', corder(kkk,:), 'DisplayName', SOLVERS_LABEL{kkk},...
        'MarkerSize', 5, 'CapSize', 20);
end

xlabel('experiments','Fontsize',12);
ylabel('correctness rate','Fontsize',12);
ylim([0.3,.85]);
xlim([.7,N_SAMPLES+.3]) 
if length(SOLVERS) ~= 1
    legend show;
else
    h_legend=legend('GEC','RGEC');
    set(h_legend,'FontSize',10, 'Location', 'NorthEast');
end

return

%figure
%for kkk = 1:length(SOLVERS)
%    nita3 = NITA{kkk}; 
%    nitb3 = NITB{kkk};
%    plot(nita3(IDX),'+','linewidth',2,'markersize',6,'color', corder(kkk,:));
%    hold on;
%    plot(nitb3(IDX),'x','linewidth',2,'markersize',6,'color', corder(kkk,:));
%end
%h_legend=legend('A plane','B plane');
%set(h_legend,'FontSize',12, 'Location', 'NorthWest');
%title('Number of DSCF iterations','FontSize',12);
%xlabel('No. experiments','Fontsize',12);
%ylabel('Total Iters.','Fontsize',12);
%return;
