% Example for robust CSP: linear mixing model classification test.

clear all 

% set NEP solvers, 'SCF' for SCF, 'simple-its' for simple iteration
%TST = 'SCF';
TST = 'simple-its';
SCFMODE = 2;    % 1: 1st order shifted; 2: 2nd order;                 

% parameters
addpath ./rcsp/
rng(0);
MISSRATE = [];
DELTA = [0.5, 1, 2, 4, 6, 8]; % perturbation levels, large value may lead to nonsmooth case 

for ii = 1:100
    [V1, lam1, Sbar1, V2, lam2, Sbar2, Tst1, Tst2, Train1, Train2] = datagen();

    % Non robust
    [V,E] = eig(Sbar2, Sbar1); 
    [~,idx] = min(diag(E));
    z0 = V(:,idx); zz1 = z0/sqrt(z0'*Sbar2*z0+z0'*Sbar1*z0);
    [~,idx] = max(diag(E));
    z0 = V(:,idx); zz2 = z0/sqrt(z0'*Sbar2*z0+z0'*Sbar1*z0);

    % Robust-delta
    for dd = 1:length(DELTA)
        delta = DELTA(dd);
        if strfind(TST, 'i')
            [z1,z2] = rbstcsp_simple(delta, Sbar1, lam1, V1, Sbar2, lam2, V2, zz1, zz2);
        else
            [z1,z2] = rbstcsp_scf(delta, Sbar1, lam1, V1, Sbar2, lam2, V2, zz1, zz2, SCFMODE);
        end
        Z1{dd} = z1;
        Z2{dd} = z2;
    end

    MissRate = csptest(zz1, zz2, Train1, Train2, Tst1, Tst2, 'Simple');

    for dd = 1:length(DELTA)
        z1 = Z1{dd}; z2 = Z2{dd};
        MissRate0 = csptest(z1, z2, Train1, Train2, Tst1, Tst2, 'SCF');
        MissRate = [MissRate; MissRate0]; 
    end
    %MissRate is a vector, the i-element records the MissRate of DELTA(i)
    MISSRATE = [MISSRATE, MissRate];
end

figure;
boxplot(1-MISSRATE')
title(TST);
ylabel('classification rate');
xticklabels(['non-rbst', string(DELTA)]);
ylim([.48,1.02])
return

%
% ****************************** AUX FUN ********************************
%

% ---------------------------------------------------- %
%   GMAT2: compute G and tilde G
% ---------------------------------------------------- %
function [G TG info] = GMAT2(w, lam2, V2, delta, Sbar1)
    [G TG] = optmat_build(w, lam2, V2, delta);
    G = Sbar1 + G;
end

% ---------------------------------------------------- %
%   HMAT2: compute H and tilde H
% ---------------------------------------------------- %
function [H TH info] = HMAT2(w, lam1, V1, delta, Sbar1)
    [H TH] = optmat_build(w, lam1, V1, delta);
    %H = Sbar1 - H;
    [H info] = NegTrunc(Sbar1 - H); % suggested in [Kawanabe, et al., 2014]
%    if ~info, disp('nonsmooth point in H matrix'), end;
    TH = -TH;
end

% ---------------------------------------------------- %
%   NegTrunc: truncate the negative eig mode in H
% ---------------------------------------------------- %
function [H issmth] = NegTrunc(H)
    [V, D] = eig(H);
    H = V*max(D,zeros(size(D)))*V';
    issmth = (min(diag(D)) >= 0);
end

% ---------------------------------------------------- %
%   rbstcsp_scf: robust csp by scf
% ---------------------------------------------------- %
function [z1,z2] = rbstcsp_scf(delta, Sbar1, lam1, V1, Sbar2, lam2, V2, zz1, zz2, mode)
    % ------------------------------
    % --- Run situation 1 
    % ------------------------------
    GMAT2d = @(w) GMAT2(w, lam2, V2, delta, Sbar2);
    HMAT2d = @(w) HMAT2(w, lam1, V1, delta, Sbar1);
    if mode == 1
        [z1, RR1, ~, ~, resd1] = scfsls(GMAT2d, HMAT2d, zz1);
    elseif mode == 2
        [z1, RR1, ~, ~, resd1] = scfls(GMAT2d, HMAT2d, zz1);
    else
        [z1, RR1, ~, ~, resd1] = scfls_sg(GMAT2d, HMAT2d, zz1);
    end
    G = GMAT2d(z1); H = HMAT2d(z1);
    z1 = z1/sqrt(z1'*G*z1+z1'*H*z1);

    % ------------------------------
    % --- Run situation 2 
    % ------------------------------
    GMAT2d = @(w) GMAT2(w, lam1, V1, delta, Sbar1);
    HMAT2d = @(w) HMAT2(w, lam2, V2, delta, Sbar2);
    if mode == 1
        [z2, RR2, ~, ~, resd2] = scfsls(GMAT2d, HMAT2d, zz2);
    elseif mode == 2
        [z2, RR2, ~, ~, resd2] = scfls(GMAT2d, HMAT2d, zz2);
    else
        [z2, RR2, ~, ~, resd2] = scfls_sg(GMAT2d, HMAT2d, zz2);
    end
    G = GMAT2d(z2); H = HMAT2d(z2);
    z2 = z2/sqrt(z2'*G*z2+z2'*H*z2);
end

% ---------------------------------------------------- %
%   rbstcsp_scf: robust csp by simple iteration
% ---------------------------------------------------- %
function [z1,z2] = rbstcsp_simple(delta, Sbar1, lam1, V1, Sbar2, lam2, V2, zz1, zz2)
    GMAT2d = @(w) GMAT2(w, lam2, V2, delta, Sbar2);
    HMAT2d = @(w) HMAT2(w, lam1, V1, delta, Sbar1);
    [z1s, RR1s, ~] = scfitr(GMAT2d, HMAT2d, zz1);
    G = GMAT2d(z1s); H = HMAT2d(z1s);
    z1 = z1s/sqrt(z1s'*G*z1s+z1s'*H*z1s);

    GMAT2d = @(w) GMAT2(w, lam1, V1, delta, Sbar1);
    HMAT2d = @(w) HMAT2(w, lam2, V2, delta, Sbar2);
    [z2s, RR2s, ~] = scfitr(GMAT2d, HMAT2d, zz2);
    G = GMAT2d(z2s); H = HMAT2d(z2s);
    z2 = z2s/sqrt(z2s'*G*z2s+z2s'*H*z2s);
end

% ---------------------------------------------------- %
%   csptest: feature extraction test 
% ---------------------------------------------------- %
function MissRate = csptest(zz1, zz2, Train1, Train2, Tst1, Tst2, FLAG)
    Tr1 = feature_ext(zz1, zz2, Train1);
    Tr2 = feature_ext(zz1, zz2, Train2);
    Ts1 = feature_ext(zz1, zz2, Tst1);
    Ts2 = feature_ext(zz1, zz2, Tst2);
    % LDA classifier
    w = lda_train(Tr1, Tr2);
    c1 = mean(Tr1'*w);
    c2 = mean(Tr2'*w);
    d1 = 1; d2 = 1;

    Tsw1 = w'*Ts1;
    NumMiss = nnz(abs(Tsw1-c1)/d1 > abs(Tsw1-c2)/d2);
    Tsw2 = w'*Ts2;
    NumMiss = NumMiss + nnz(abs(Tsw2-c2)/d2 > abs(Tsw2-c1)/d1);
    MissRate = NumMiss/(length(Tsw1)+length(Tsw2));
    disp([FLAG,':   miss = ', num2str(NumMiss),'  missrate = ',num2str(MissRate)]);
end
