function [w, optr] = rlda_cvx(Sx, Sy, dx, dy, ux, uy, Px, Py)
% Solve RLDA via QCQP with 2 inequalities:
%           minimize    (x-y)'*P0^{-1}*(x-y)
%               s.t.    (x-ux)'*Px*(x-ux) - 1 <= 0  
%                       (y-uy)'*Py*(y-uy) - 1 <= 0  
% and verifies that strong duality holds.
%
% There are 4 different approaches to chose from.

n = length(Sx);
P0 = Sx+Sy+(dx+dy)*eye(n);

% ---- approach I ---
%PP = [eye(n), -eye(n)]; 
%PP0 = PP'*(P0\PP);
%PPx = blkdiag(Px, zeros(n));
%PPy = blkdiag(zeros(n), Py);
%q1 = -2*[Px*ux; zeros(n,1)]; r1 = ux'*Px*ux -1;
%q2 = -2*[zeros(n,1); Py*uy]; r2 = uy'*Py*uy -1;
%cvx_begin quiet
%    cvx_precision high
%    variable x(2*n)
%    dual variables lam1 lam2 
%    minimize( quad_form(x,PP0))
%    lam1: quad_form(x,PPx) + q1'*x + r1 <= 0;
%    lam2: quad_form(x,PPy) + q2'*x + r2 <= 0;
%cvx_end
%w = P0\(x(1:n) - x(n+1:end));

% ---- approach II ---
IP0 = pinv(P0);
q1 = -2*Px*ux; r1 = ux'*Px*ux -1;
q2 = -2*Py*uy; r2 = uy'*Py*uy -1;
cvx_begin quiet
    cvx_precision high
    variables x(n) y(n);
    dual variables lam1 lam2 
    minimize( quad_form(x-y,IP0))
    lam1: quad_form(x,Px) + q1'*x + r1 <= 0;
    lam2: quad_form(y,Py) + q2'*y + r2 <= 0;
cvx_end
w = IP0*(x - y);

% ---- approach III ---
%L = chol(P0, 'lower');
%q1 = -2*L'*Px*ux; r1 = ux'*Px*ux -1;
%q2 = -2*L'*Py*uy; r2 = uy'*Py*uy -1;
%Px = L'*Px*L;
%Py = L'*Px*L;
%cvx_begin quiet
%    cvx_precision high
%    variables x(n) y(n);
%    dual variables lam1 lam2 
%    minimize( norm(x-y) )
%    lam1: quad_form(x,Px) + q1'*x + r1 <= 0;
%    lam2: quad_form(y,Py) + q2'*y + r2 <= 0;
%cvx_end
%w = L'\(x - y);
%

% ---- approach IV ---
%IP0 = pinv(P0);
%q1 = -2*Px*ux; r1 = ux'*Px*ux -1;
%q2 = -2*Py*uy; r2 = uy'*Py*uy -1;
%cvx_begin quiet
%    cvx_precision high
%    variables x(n) y(n);
%    minimize( quad_form(x-y,IP0))
%    subject to
%        quad_form(x,Px) + q1'*x + r1 <= 0;
%        quad_form(y,Py) + q2'*y + r2 <= 0;
%cvx_end
%w = IP0*(x - y);


obj1 = cvx_optval;
optr = 1/obj1;
return




