function [z1, r, R, its, RESD] = scfls(EvalG, EvalH, z1, dolsearch, tol)
% SCF iteration applied to the second order NEP with/without linesearch.
%
% INPUT:
%   EvalG   function handle 
%           G(z) = EvalG(z)
%           [G(z), TG(z)] = EvalG(z)
%   EvalH   function handle 
%           H(z) = EvalH(z)
%           [H(z), TH(z)] = EvalG(z)
%   z1      initial vector
%   dolsearch do line search if true (default true)
%   tol     tolerance 
%
% OUTPTU:
%   z1:     initial guess
%   r :     minimum ratio rho
%   R :     each column contains the computed z from each iteration.
%   its:    number of iterations
%   RESD:   relative residual norm for the 1st NEP of each iteration.
%

UNIVARSEARCH = false; % DISALLOWED: set TRUE to use interpolation

if nargin < 4
    dolsearch = true;
end
if nargin < 5
    tol   = 1.0E-8; 
end
maxit = 200;
R = [];
r = [];
RESD = []; 

% +++++++++++++++++++++++++++
% --- MAIN LOOP
% +++++++++++++++++++++++++++
jj = 0;
Nls = 0; % number of line search
for i = 1: maxit
    % Evaluate ratio: lsearch mode can reuse the matrices
    [AA,TG] = EvalG(z1);
    [BB,TH] = EvalH(z1);
    rr = z1'*AA*z1 / (z1'*BB*z1);     
    r = [r,rr];
    R = [R,z1];

    % relative residual norm
    resd = norm(AA*z1-rr*BB*z1)/(norm(AA) + rr*norm(BB))/norm(z1);
    RESD = [RESD, resd];

    % Stopping cnd
    if i > 1 && resd < tol %abs(rr - r(end-1)) < tol
        break;
    end

    % SCF updating
    [V,E] = eig(AA+TG, BB+TH);
    tt = diag(E);
    IDX = find(tt==min(tt(tt>0)));
    IDX = IDX(end);

    z0 = z1; 
    z10 = V(:,IDX); z10 = z10/norm(z10);
    if rr~= tt(IDX) 
        z10 = z10*sign(z10'*BB*z0)*sign(rr - tt(IDX));
    end
    sd = -2*(AA*z1 - rr*BB*z1)/(z1'*BB*z1); % steepest ascend

    if dolsearch 
        % evaluate new ratio to determine
        AA = EvalG(z10);
        BB = EvalH(z10);
        rrt = z10'*AA*z10 / (z10'*BB*z10);     

        % check sufficient reduction: Armijo
        if rrt > rr - abs(sd'*(z10-z0))*0.01
            if UNIVARSEARCH
                d = z10-z0;
                % cubic interpolation, and optimizer...
                %	f0 = rho(x_k) 
                %	f1 = rho(x_{k+1}) 
                % 	f'(0) = (x_{k+1} - x_k)'*g(x_k) 
                % 	f'(1) = (x_{k+1} - x_k)'*g(x_{k+1}) 
                %	x_{k+1} = x_k + alpha*(x_{k+1}-x_k) 
                sd1 = -2*(AA*z10 - rrt*BB*z10)/(z10'*BB*z10); 
                alpha = univar_search(rr, rrt, -d'*sd, -d'*sd1);
                z1t = z0 + alpha*d;
                lsinfo = true;
            else
                d = z10-z0;
                d = d - z0*z0'*d;
                alpha = d'*sd/norm(sd)/norm(d);
                if abs(alpha) < 0.01
                    d = d/norm(d) + 0.01*sd/norm(sd);
                end
                d = d/norm(d);
                mm = d'*sd;
                cc = 0.01;
                Nls = Nls + 1;
                gamma = norm(z10-z0);
                lsinfo = false;
                while gamma>1.0E-15 
                    z1t = z0 + gamma*d;
                    AA = EvalG(z1t);
                    BB = EvalH(z1t);
                    rrt = z1t'*AA*z1t / (z1t'*BB*z1t);     
                    if rrt <rr - gamma*abs(mm)*cc
                    	lsinfo = true;
                    	break
                    end
                    gamma = gamma/10;
                    jj = jj+1;
                end
            end
        else
            z1t = z10;
            lsinfo = true;
        end 

        if ~lsinfo
            % line search fail, possibly due to non-smoothness or numerical error 
            [rr,IDXX] = min(r);
            z1 = R(:,IDXX);
            AA = EvalG(z1);
            BB = EvalH(z1);
            resd = norm(AA*z1-rr*BB*z1)/(norm(AA) + rr*norm(BB))/norm(z1);
            break; 
        end
        z1 = z1t;
    else
        z1 = z10;
    end
    z1 = z1/norm(z1);
end
its = i + jj; % inner outer its sum
return;
