function [z1, r, R, its, RESD] = scfsls(EvalG, EvalH, z1, dolsearch, tol)
% SCF iteration applied to the 1st order shifted NEP with/without linesearch.
%
% INPUT:
%   EvalG   function handle G(z) = EvalG(z)
%   EvalH   function handle H(z) = EvalH(z)
%   z1      initial vector
%   dolsearch do lin search if true (default true)
%   tol     tolerance
%
% OUTPTU:
%   z1:     initial guess
%   r :     minimum ratio rho
%   R :     each column contains the computed z from each iteration.
%   its:    number of iterations
%   RESD:   relative residual norm for the 1st NEP of each iteration.
%

UNIVARSEARCH = false; % DISALLOWED: set TRUE to use interpolation

mode = 1; % EXPERIMENTAL: 0 = auto switching mode; 1 = always run shifting mode
sigma = 0; 
beta = 1.01; 

if nargin < 4
    dolsearch = true;
end
maxit = 200;
if nargin < 5 
    tol   = 1.0E-6; 
end

R = [];
r = [];
RESD = [];
 
% +++++++++++++++++++++++++++
% --- MAIN LOOP 
% +++++++++++++++++++++++++++
jj = 0;
Nls = 0; % number of linesearch
for i = 1: maxit
    % Evaluate ratio: lsearch mode can reuse the matrices
    if i == 1 || ~dolsearch || mode~=1
        AA = EvalG(z1);
        BB = EvalH(z1);
    end
    rr = z1'*AA*z1 / (z1'*BB*z1);     

    if mode == 0 && i>1 && (r(end) < rr);
        mode = 1;
        z1 = R(:,end);
        AA = EvalG(z1);
        BB = EvalH(z1);
        rr = r(end);
    else
        r = [r,rr];
        R = [R,z1];
        resd = norm(AA*z1-rr*BB*z1)/(norm(AA*z1) + rr*norm(BB*z1));
        RESD = [RESD, resd];
    end

    % Stopping cnd
    if i > 1 && resd < tol %abs(rr - r(end-1)) < tol
        break;
    end

    % SCF to shifted 1st NEP
    [V,E] = eig(AA, BB);
    tt = diag(E);
    tt = tt(~isinf(tt));
    if mode == 1
        sigma = beta*max(tt) - min(tt);
        BBz = BB*z1; BBz=BBz/sqrt(z1'*BBz);
        [V, E] = eig(AA-sigma*BBz*BBz', BB);
    end

    [~, IDX] = min(diag(E));
    z0 = z1; % record original z
    z10 = V(:,IDX); z10 = z10/norm(z10);

    ssign = sign(z10'*BB*z0)*sign(rr - sigma- E(IDX,IDX));
    if ssign == 0 break;end

    z10 = z10*ssign;
    sd = -2*(AA*z1 - rr*BB*z1)/(z1'*BB*z1); % steepest ascend

    if mode==1 && dolsearch 
        % evaluate new ratio to determine
        AA = EvalG(z10);
        BB = EvalH(z10);
        rrt = z10'*AA*z10 / (z10'*BB*z10);     

        % check sufficient reduction: Armijo
        if rrt > rr - abs(sd'*(z10-z0))*0.01
            if UNIVARSEARCH
                d = z10-z0;
                % cubic interpolation, and optimizer...
                %   f0 = rho(x_k) 
                %   f1 = rho(x_{k+1}) 
                %   f'(0) = (x_{k+1} - x_k)'*g(x_k) 
                %   f'(1) = (x_{k+1} - x_k)'*g(x_{k+1}) 
                %   x_{k+1} = x_k + alpha*(x_{k+1}-x_k) 
                sd1 = -2*(AA*z10 - rrt*BB*z10)/(z10'*BB*z10); 
                alpha = univar_search(rr, rrt, -d'*sd, -d'*sd1);
                z1t = z0 + alpha*d;
                AA = EvalG(z1t);
                BB = EvalH(z1t);
            else
                d = z10-z0;
                d = d - z0*z0'*d;
                alpha = d'*sd/norm(sd)/norm(d);
                if abs(alpha) < 0.01
                    d = d/norm(d) + 0.01*sd/norm(sd);
                end
                d = d/norm(d);
                
                mm = d'*sd;
                cc = 0.01;
                Nls = Nls + 1;
                
                gamma = norm(z10-z0);
                lsinfo = false;
                while gamma>1.0E-15 
                    z1t = z0 + gamma*d;
                    AA = EvalG(z1t);
                    BB = EvalH(z1t);
                    rrt = z1t'*AA*z1t / (z1t'*BB*z1t);     
                    if rrt <rr - gamma*abs(mm)*cc
                        lsinfo = true;
                        break
                    end
                    gamma = gamma/10;
                    jj = jj+1;
               end
            end
        else
            z1t = z10;
            lsinfo = true;
        end 

        if ~lsinfo
            % line search fail, possibly due to non-smoothness or numerical error 
            [rr,IDXX] = min(r);
            z1 = R(:,IDXX);
            AA = EvalG(z1);
            BB = EvalH(z1);
            resd = norm(AA*z1-rr*BB*z1)/(norm(AA) + rr*norm(BB))/norm(z1);
            break; 
        end

        z1 = z1t;
    else
        z1 = z10;
    end
    z1 = z1/norm(z1);
end
its = i + jj; % inner outer its sum
return;
